% Complex numbers, mostly from http://www.trapexit.org/Complex_Numbers

-module(complex).
-export([make/1, make/2, real/1, imaginary/1, is_complex/1, 
         add/2, sub/2, mult/2, divide/2]).

-record(complex, {real, imaginary}).

%: +type complex() = {complex, float(), float()}.

%: +func real :: complex() -> float().
real(X) -> X#complex.real.

%: +func imaginary :: complex() -> float().
imaginary(X) -> X#complex.imaginary.

%: +func is_complex :: any() -> erlang:bool().
is_complex(X) -> is_record(X, complex).


%: +func make :: float() * float() -> complex().
make(Real, Imaginary) -> #complex{real = Real, imaginary = Imaginary}.

%: +func make :: (float() + complex()) -> complex().
make(X) when is_tuple(X) -> X;
%make(X) when is_record(X, complex) -> X;
make(Real) -> make(Real, 0).


%: +func add :: complex() * complex() -> complex().
add(X, Y) ->
  A = make(X), 
  B = make(Y),
  make(real(A) + real(B), imaginary(A) + imaginary(B)).

%: +func sub :: complex() * complex() -> complex().
sub(X, Y) ->
  A = make(X), 
  B = make(Y),
  make(real(A) - real(A), imaginary(B) - imaginary(B)).

%: +func mult :: complex() * complex() -> complex().
mult(X, Y) ->
  A = make(X), 
  B = make(Y),
  R = (real(A) * real(B)) - (imaginary(A) * imaginary(B)),
  I = (real(A) * imaginary(B)) + (real(B) * imaginary(A)),
  make(R,I).

%: +func divide :: complex() * complex() -> complex().
divide(X,Y) ->
  A = make(X), 
  B = make(Y),
  Divisor = math:pow(real(B),2) + math:pow(imaginary(B),2),
  R = ((real(A) * real(B)) + (imaginary(A) * imaginary(B))) / Divisor,
  I = ((imaginary(A) * real(B)) - (real(B) * imaginary(B))) / Divisor,
  make(R,I).
