-module(mandelbrot).
-export([start/1, value_at/3, loop/1]).


%: +func sqr_len :: complex:complex() -> float().
sqr_len(C) ->
  R = complex:real(C),
  I = complex:imaginary(C),
  R * R + I * I.

%: +func mandel :: complex:complex() * integer() * complex:complex() * integer() 
%:              -> integer().
mandel(C,MaxI,Z,I) ->
  case (I < MaxI) and (sqr_len(Z) < 2.0) of
    true ->
      Zsqrd = complex:mult(Z,Z),
      Znext = complex:add(C,Zsqrd),
      mandel(C,MaxI,Znext,I+1);
    false ->
      I
  end.

%: +func start :: integer() -> pid().
start(MaxI) -> spawn(mandelbrot,loop,[MaxI]).

%: +func value_at :: pid() * float() * float() -> float().
value_at(Pid,X,Y) -> 
  rpc(Pid,{mandel,X,Y}).

rpc(Pid,Req) ->
  Pid ! {self(), Req},
  receive
    {Pid, Resp} -> 
      Resp
  end.

%: +func loop :: integer() -> none().
loop(MaxIter) ->
  receive
    {From,{mandel,X0,Y0}} ->
      X = X0, 
%:    promise X :: float(),
      Y = Y0,
%:    promise Y :: float(),
      C = complex:make(X,Y),
      V = mandel(C,MaxIter,C,1),
      From ! {self(),V},
      loop(MaxIter);
    _ ->
      io:format("What?~n"),
      loop(MaxIter)
  end.
