%%% @author Sven-Olof Nystrom <svenolof@hyena>
%%% @copyright (C) 2015, Sven-Olof Nystrom
%%% @doc
%%%
%%% @end
%%% Created : 10 Aug 2015 by Sven-Olof Nystrom <svenolof@hyena>
 
-module(conn).

%-compile(export_all).
-export([any/0, is_any/1]).
-export([con_ordered/2, con_included/2, variance/2, match/2]).
-export([arg/0, pid/0, port/0, reference/0, nil/0, atom/0, atom_fixed/1]).
-export([bitstring/0, float/0, fun_any/0, fun_fixed/1, integer/0, cons/0]).
-export([tuple/0, tuple_fixed/1, tuple_tagged/2, map/0, set/0, abstract/1]).

%: +opaque con() = arg + pid + port + reference + nil 
%:               + atom + {atom_fixed, atom()} + bitstring 
%:               + float + fun_any + {fun_fixed, integer()}
%:               + integer + cons 
%:               + tuple + {tuple_fixed, integer()} 
%:               + {tuple_tagged, atom(), integer()} 
%:               + map + set
%:               + {abstract, name()} + '$any'.

%: +type name() = any().

%: +type texp() =  walker:texp0().
%: +type constraint() = solver:constraint().
%: +type list(X) = erlang:list(X).

%: +func arg :: () -> con().
arg() ->
    arg.

%: +func pid :: () -> con().
pid() ->
    pid.

%: +func port :: () -> con().
port() ->
    port.

%: +func reference :: () -> con().
reference() ->
    reference.

%: +func nil :: () -> con().
nil() ->
    nil.

%: +func atom :: () -> con().
atom() ->
    atom.

%: +func atom_fixed :: atom() -> con().
atom_fixed(A) ->
    {atom_fixed, A}.

%: +func bitstring :: () -> con().
bitstring() ->
    bitstring.
    
%: +func float :: () -> con().
float() ->
    float.

%: +func fun_any :: () -> con().
fun_any() ->
   fun_any.

%: +func fun_fixed :: integer() -> con().
fun_fixed(N) ->
    {fun_fixed, N}.

%: +func integer :: () -> con().
integer() ->
    integer.

%: +func cons :: () -> con().
cons() ->
    cons.

%: +func tuple :: () -> con().
tuple() ->
    tuple.

%: +func tuple_fixed :: integer() -> con().
tuple_fixed(N) ->
    {tuple_fixed, N}.

%: +func tuple_tagged :: atom() * integer() -> con().
tuple_tagged(A, N) ->
    {tuple_tagged, A, N}.

%: +func map :: () -> con().
map() ->
    map.

%: +func set :: () -> con().
set() ->
    set.


%: +func abstract :: any() -> con().
abstract(N) ->
    {abstract, N}.

%: +func any :: () -> con().
any() ->
    '$any'.

%: +func is_any :: con() -> true + false.
is_any('$any') ->
    true;
is_any(_) ->
    false.


% Total order of constructors. 
%: +func con_ordered :: con() * con() -> true + false.
con_ordered(C1, C2) when is_atom(C1), is_atom(C2) ->
    C1 =< C2;
con_ordered(C, _) when is_atom(C) ->
    true;
con_ordered(_, C) when is_atom(C) ->
    false;
con_ordered({atom_fixed, A1}, {atom_fixed, A2}) ->
    A1 =< A2;
con_ordered({atom_fixed, _}, _) ->
    true;
con_ordered(_, {atom_fixed, _}) ->
    false;
con_ordered({fun_fixed, N1}, {fun_fixed, N2}) ->
    N1 =< N2;
con_ordered({fun_fixed, _}, _) ->
    true;
con_ordered(_, {fun_fixed, _}) ->
    false;
con_ordered({tuple_fixed, N1}, {tuple_fixed, N2}) ->
    N1 =< N2;

con_ordered({tuple_fixed, N1}, {tuple_tagged, _, N2}) ->
    N1 =< N2;
con_ordered({tuple_tagged, _, N1}, {tuple_fixed, N2}) ->
    N1 < N2;

con_ordered({tuple_fixed, _}, _) ->
    true;
con_ordered(_, {tuple_fixed, _}) ->
    false;
con_ordered({tuple_tagged, A1, _N1}, {tuple_tagged, A2, _N2}) when A1 < A2 ->
    true;
con_ordered({tuple_tagged, A1, N1}, {tuple_tagged, A2, N2}) when A1 == A2 ->
    N1 =< N2;
con_ordered({tuple_tagged, _, _}, {tuple_tagged, _, _})  ->
    false;    
con_ordered({tuple_tagged, _, _}, _)  ->
    true;
con_ordered(_, {tuple_tagged, _, _})  ->
    false;        

con_ordered({abstract, F1}, {abstract, F2}) ->
    F1 =< F2;
con_ordered({abstract, _}, _) ->
    true;
con_ordered(_, {abstract, _}) ->
    false.

%: +func variance :: con() * list(X) -> {list(X), list(X)}.
variance({fun_fixed, _}, [E|EL]) ->
    {[E], EL};    
variance(_, L) ->
    {L, []}.

% Inclusion order of constructors. 
%: +func con_included :: con()*con() -> true + false.
con_included(C, C) ->
    true;
con_included(_, '$any') ->
    true;
con_included(integer, float) ->
    true;
con_included(C, _) when is_atom(C) ->
    false;
con_included({atom_fixed, _}, atom) ->
    true;
con_included({fun_fixed, _}, fun_any) ->
    true;
con_included({tuple_fixed, _}, tuple) ->
    true;
con_included({tuple_tagged, _, _}, tuple) ->
    true;
con_included({tuple_tagged, _, N}, {tuple_fixed, N}) ->
    true;
con_included(_, _) ->
    false.




%: +func match :: texp() * texp() -> 
%:                none + {value,  list({include, texp(), texp()})}.
match({c, pid, []}, {c, pid, []}) ->
    {value, []};

match({c, port, []}, {c, port, []}) ->
    {value, []};

match({c, reference, []}, {c, reference, []}) ->
    {value, []};

match({c, nil, []}, {c, nil, []}) ->
    {value, []};



match({c, atom, []}, {c, atom, []}) ->
    {value, []};
match({c, {atom_fixed, _A}, []}, {c, atom, []}) ->
    {value, []};
match({c, {atom_fixed, A}, []}, {c, {atom_fixed, A}, []}) ->
    {value, []};

match({c, bitstring, []}, {c, bitstring, []}) ->
    {value, []};

match({c, float, []}, {c, float, []}) ->
    {value, []};

match({c, fun_any, [T]}, {c, fun_any, [U]}) ->
    {value, [{include, T, U}]};
match({c, {fun_fixed, _N}, TList},{c, fun_any, [U]}) ->
    {value, [{include, T, U}||T<-TList]};
match({c, {fun_fixed, N}, [T|TList]}, {c, {fun_fixed, N}, [U|UList]}) ->
    {value, [{include, T,U}]++include_lists(UList, TList)};

match({c, integer, []}, {c, integer, []}) ->
    {value, []};
%% match({c, integer, []}, {c, float, []}) ->
%%     {value, []};

match({c, cons, [T1, T2]}, {c, cons, [U1, U2]}) ->
    {value, [{include, T1, U1},
	     {include, T2, U2}]};

match({c, tuple, [T1]}, {c, tuple, [U1]}) ->
    {value, [{include, T1, U1}]};
match({c, tuple, [T1]}, {c, {tuple_tagged, _A, _}, UList}) ->
    {value, [{include, T1, U}|| U <- UList]};
match({c, tuple, [T1, T2]}, U1 = {c, {tuple_fixed, _}, UList}) ->
    {value, [{include, T1, U1}]++
            [{include, T2, U}|| U <- UList]};

match({c, {tuple_fixed, _}, TList}, {c, tuple, [U1]}) ->
    {value, [{include, T, U1} || T <- TList]};
match({c, {tuple_fixed, N}, TList}, {c, {tuple_fixed, N}, UList}) ->
    {value, include_lists(TList, UList)};
match({c, {tuple_fixed, N}, [_T|TList]}, {c, {tuple_tagged, _A, N}, UList}) ->
    {value, include_lists(TList, UList)};

match({c, {tuple_tagged, A, _}, TList}, {c, tuple, [U1]}) ->
    {value, [{include, {c, {atom_fixed, A}, []}, U1}] ++
	    [{include, T, U1} || T <- TList]};
match({c, {tuple_tagged, A, N}, TList}, {c, {tuple_fixed, N}, [U|UList]}) ->
    {value, [{include, {c, {atom_fixed, A}, []}, U}]++
         include_lists(TList, UList)};
match({c, {tuple_tagged, A, N}, TList}, {c, {tuple_tagged, A, N}, UList}) ->
    {value, include_lists(TList, UList)};

match({c, map, [T1, T2]}, {c, map, [U1, U2]}) ->
    {value, [{include, T1, U1},
    	     {include, T2, U2}]};

match({c, {abstract, F}, TList}, {c, {abstract, F}, UList}) ->   
    {value, include_lists(TList, UList)};

match({c, '$any', []}, {c, _, TList}) ->
     {value, [{include, {c, '$any', []}, T} || T <- TList]};

match(_A, _B) -> 
    none. % Note that this does not imply a fail.

include_lists([], []) ->
    [];
include_lists([T|TList],[U|UList]) ->
    [{include, T, U}|include_lists(TList, UList)].
    

%% con(pid)                  -> pid;                     % 0
%% con(port)                 -> port;                    % 0
%% con(reference)            -> reference;               % 0
%% con(nil)                  -> nil;                     % 0
%% con(atom)                 -> atom;                    % 0
%% con({atom_fixed, A})      -> {atom_fixed, A};         % 0
%% con(bitstring)            -> bitstring;               % 0
%% con(float)                -> float;                   % 0
%% con(fun_any)              -> fun_any;                 % 1
%% con({fun_fixed, N})       -> {fun_fixed, N};          % N+1 
%% % First parameter is return type.
%% con(integer)              -> integer;                 % 0
%% con(cons)                 -> cons;                    % 2
%% con(tuple)                -> tuple;                   % 1
%% con({tuple_fixed, N})     -> {tuple_fixed, N};        % N
%% con({tuple_tagged, A, N}) -> {tuple_tagged, A, N};    % N -1
%% con({abstract, F})        -> {abstract, F}.           % ?

