-module(display).

-compile(export_all).



%%%% ------------------------------------------------- %%%%

-export([predicates/1]).
%: +func predicates :: gb_trees:tree(any(), walker:predicate()) 
%:                 -> prettypr:document().                                          


predicates(Predicates) ->
    Docs = 
	lists:map(fun(Predicate) ->
			  predicate(Predicate)
		  end,
		  gb_trees:values(Predicates)),
    prettypr:par(Docs).

predicate_list(L) ->
    Docs = 
	lists:map(fun(Predicate) ->
			  predicate(Predicate)
		  end,
		  L),
    prettypr:par(Docs).
    

predicate({predicate, Name, ArgList, Body}) ->
    ConstraintList = lists:flatten(Body),
    prettypr:above(prettypr:sep([prettypr:text("predicate"),
                                 name(Name),
                                 arg_list(ArgList)]),
                    prettypr:nest(4, constraint_list(ConstraintList))).

name(Name) when is_atom(Name) ->
    prettypr:text(io_lib:format("~s", [atom_to_list(Name)]));
name(Name) ->
    prettypr:text(io_lib:format("~w", [Name])).

    
arg_list(ArgList) ->
    case [name(Arg) || Arg <- ArgList] of
	[PArg|PArgList] ->
	    prettypr:beside(
	      prettypr:sep([prettypr:beside(
			      prettypr:text("["),
			      PArg)]
			   ++ PArgList),
	      prettypr:text("]"));
	[] ->
	    prettypr:text("[]")
    end.
	    


constraint_list([]) ->
    prettypr:text("true");
constraint_list([C]) ->
    prettypr:beside(constraint(C), prettypr:text("."));
constraint_list([C|Cl]) ->
    prettypr:above(prettypr:beside(constraint(C), prettypr:text(",")),
                  constraint_list(Cl)).

constraint({call, Name, NameList}) ->
    prettypr:sep([prettypr:text("call"),
		 name(Name),
                  arg_list(NameList)]);
constraint({include, Exp1, Exp2, _}) ->
    prettypr:sep([exp(Exp1), prettypr:text("≤"), exp(Exp2)]);
constraint({watch, Name, String}) ->
        prettypr:sep([
                      prettypr:text("watch"),
                      name(Name),
                      prettypr:text("\""),
                      prettypr:text(String),
                      prettypr:text("\"")                      
                     ]).


exp({var, Name}) ->
    name(Name);
exp({sift, Name, Filter}) ->
    prettypr:sep([
                  name(Name), 
                  prettypr:text("$"), 
                  filter(Filter)
                 ]);
exp({trim, Name, C}) ->
    prettypr:sep([
                  name(Name), 
                  prettypr:text("$$"), 
                  con(C)
                 ]);
exp({c, C, ExpList}) ->
    prettypr:beside(
      prettypr:sep([prettypr:beside(
                      prettypr:text("〈"),
                      con(C))]
                   ++ lists:map(fun exp/1, ExpList)),
      prettypr:text("〉")).

% Please note that this is specific to rfilter.

%: +type filter() = list({filter:con(), list(filter:con())}).

%% So [{C1, []}, {C2, [D1, D2]}] represents
%%     C1 ∪ (C2 \ (D1 ∪ D2))

filter([]) ->
    prettypr:text("0");
filter([Clause]) ->
    filter_clause(Clause);
filter([Clause|ClauseList]) ->
    prettypr:sep([
                  filter_clause(Clause),
                  prettypr:text("∪"),
                  filter(ClauseList)
                  ]).

filter_clause({C, []}) ->
    con(C);
filter_clause({C, [C1]}) ->
    prettypr:sep([
                  con(C),
                  prettypr:text("\\"),
                  con(C1)
                 ]);   
filter_clause({C, CL}) ->
    prettypr:sep([
                  con(C),
                  prettypr:text("\\"),
                  exclude_list(CL)
                 ]).

exclude_list([C]) ->
    con(C);
exclude_list([C|CL]) ->
    prettypr:sep([
                  con(C),
                  prettypr:text("∪"),
                  exclude_list(CL)
                 ]).

con({abstract, X}) when is_atom(X) ->
    prettypr:text(io_lib:format("& ~s", [atom_to_list(X)]));
con({abstract, X}) ->
    prettypr:text(io_lib:format("ABSTRACT ~w", [X]));
con({fun_fixed, N}) ->
    prettypr:text(io_lib:format("→/~w", [N]));
con(Name) ->
    prettypr:text(io_lib:format("~w", [Name])).





