-module(match).

-export([match/3, match_head/4]).
% match
%: +type list(X) = erlang:list(X).

%: +type epattern() = 
%:       {unfilter, list(walker:con()), epattern()}
%:     + {empty, list(walker:name())} 
%:     + {c, con(), list(epattern()), list(walker:name())}
%:     + unmatchable.

%: +type pattern() = convert:pattern().
%: +type con() = conn:con().

%: +func match_head :: list(epattern()) 
%:                   * list(walker:texp0()) 
%:                   * list(list(epattern()))
%:                   * walker:pos() 
%:             ->  {!walker:constraints(), !walker:bind()} + unmatchable.
match_head(PHead, THead, QList,  Pos) ->
    PHead1 = trim_head(PHead, QList),
    match_list(PHead1, THead, Pos).

%: +func match :: epattern() * walker:texp0() * walker:pos() 
%:             ->  {!walker:constraints(), !walker:bind()} + unmatchable.
match({empty, VarList}, T, _Pos) ->
    B = bind_var_list(VarList, T),
    {[], B};
match({c, C, PList, VarList}, T, Pos) ->
    Source = walker:source_empty(),
    Temp = name_new(pos:left(pos:left(Pos))),
    TempList = temp_new(PList, pos:right(pos:left(Pos))),
    G1 = {include, 
	  walker:apply_filter(T, rfilter:singleton( C)), {var, Temp}, Source},
    G2 = {include, {var, Temp}, {c, C, TempList}, Source},
    {G3, B1} = match_list(PList, TempList, pos:right(Pos)),
    B2 = bind_var_list(VarList, {var, Temp}),
    {[G1, G2, G3], walker:bind_combine(B1, B2)};
match({unfilter, CList, P}, T, Pos) ->
    Source = walker:source_empty(),
    F1 = unfilter_list(CList),
    T1 = name_new(pos:left(Pos)),
    G = {include, walker:apply_filter(T, F1), {var, T1}, Source},
    {G1, B1} = match(P, {var, T1}, pos:right(Pos)),
    {[G | G1], B1};
match(unmatchable, _, _) ->
    unmatchable.



%: +func unfilter_list :: list(walker:con()) -> rfilter:filter().
unfilter_list([]) ->
    rfilter:any();
unfilter_list([C|L]) ->
    rfilter:intersect(rfilter:singleton_inverse(C), unfilter_list(L)).


match_list([], [], _) ->
    {[], walker:bind_empty()};
match_list([Pattern|PatternList], [T|TList], Pos) ->
    case match(Pattern, T,  pos:left(Pos)) of
        {G1, B1} -> 
            {G2, B2} = match_list(PatternList, TList, pos:right(Pos)),
            B = walker:bind_combine(B1, B2),
            {[G1 | G2], B};
        unmatchable ->
            unmatchable
    end;
match_list(unmatchable, _, _) ->
    unmatchable.


bind_var_list([Var|VarList], T) ->
    B1 = walker:bind_singleton(Var, T),
    B2 = bind_var_list(VarList, T),
    walker:bind_combine(B1, B2);
bind_var_list([], _) ->
    walker:bind_empty().

%% % Trim
%: +func trim_head :: (unmatchable+list(epattern())) * list(list(epattern()))
%:                 -> (unmatchable+list(epattern())).
trim_head(PArg, []) ->
    PArg;
trim_head(unmatchable, _) ->
    unmatchable;
trim_head(PArg, [QArg|QArgList]) ->
    case antimatches_head(PArg, QArg) of
	[AM] ->
	    PArg1 = add_filter_head(PArg, AM);
	[] ->
	    PArg1 = unmatchable;

	[_|_] ->
	    PArg1 = PArg
    end,
    trim_head(PArg1, QArgList).

% antimatches
%: +func antimatches_head :: list(epattern()) * list(epattern()) 
%:                   -> list({place(), con()}).
antimatches_head(PList, QList) -> 
    PlaceList=make_place_list([], conn:arg(), 0, PList), 
    antimatches_list(PlaceList, PList, QList).

%: +func antimatches :: place() * epattern() * epattern() -> list({place(), con()}).

antimatches(_, {empty, _}, {empty, _}) ->
    [];
antimatches(_, {c, _, _, _}, {empty, _}) ->
    [];
antimatches(Place, {unfilter, _CList, P}, Q) ->
    antimatches(Place, P, Q);

antimatches(Place, {empty, _}, {c, C, QList, _}) ->
    PList = make_dummy_list(QList),
    PlaceList = make_place_list(Place, C, QList),
    [{lists:reverse(Place), C}] ++ antimatches_list(PlaceList, PList, QList);
antimatches(Place, {c, C, _, _}, {c, C1, QList, _}) when C =/= C1 ->
    PList = make_dummy_list(QList),
    PlaceList = make_place_list(Place, C, QList),
    [{lists:reverse(Place), C1}] ++ antimatches_list(PlaceList, PList, QList);
antimatches(Place, {c, C, PList, _}, {c, C, QList, _}) ->
    PlaceList = make_place_list(Place, C, PList),
    antimatches_list(PlaceList, PList, QList).

%: +func antimatches_list :: list(place()) 
%:                         * list(epattern())
%:                         * list(epattern()) 
%:                        -> list({place(), con()}).
antimatches_list([], [], []) ->
    [];
antimatches_list([Place|PlaceList], [P|PList], [Q|QList]) ->
    antimatches(Place, P, Q) ++ 
	antimatches_list(PlaceList, PList, QList).

make_dummy_list([]) ->
    [];
make_dummy_list([_|L]) ->
    [{empty, []}|make_dummy_list(L)].


%: +func make_place_list :: place() * con() * list(any()) ->  list(place()).
make_place_list(Place, C, L) ->
     make_place_list(Place, C, 0, L).

make_place_list(_, _, _, []) ->
    [];
make_place_list(Place, C, I, [_|L]) ->
    [[{C, I}|Place] | make_place_list(Place, C, I+1, L)].

%% Add filter

%: +func add_filter_head :: list(epattern()) * {place(), con()} 
%:                       -> list(epattern()).
add_filter_head(PArg, F) ->
    {c, _, PArg1, []} = add_filter({c, conn:arg(), PArg, []}, F),
    PArg1.

%% First clause seems redundant?
%: +type place() = list({con(), integer()}).
%: +func add_filter :: epattern() * {place(), con()} -> epattern().

%% add_filter({empty, VarList}, {[], C}) -> 
%%     {unfilter, [C], {empty, VarList}};
%% add_filter({unfilter, L,  P}, {[], C}) -> 
%%    {unfilter, [C|L], P};
add_filter(P, {[], C}) -> 
   {unfilter, [C], P};

add_filter({c, C, PList, VarList}, {[{C, I}|Place], C1}) ->
    PList1 = add_filter_indexed(PList, 0, I, {Place, C1}),
    {c, C, PList1, VarList};
add_filter({unfilter, L, P}, {Place, C}) ->
    P1 = add_filter(P, {Place, C}),
    {unfilter, L, P1}.


add_filter_indexed([], _, _, _) ->
    [];
add_filter_indexed([P|PList], J, I, {Place, C}) when J==I ->
    P1 = add_filter(P, {Place, C}),			      		       
    PList1 = add_filter_indexed(PList, J+1, I, {Place, C}),
    [P1|PList1];
add_filter_indexed([P|PList], J, I, {Place, C}) when J=/=I ->
    PList1 = add_filter_indexed(PList, J+1, I, {Place, C}),
    [P|PList1].

%%%%

make_name(P, N) ->
    String = lists:flatten(io_lib:format("~s~w", [P, N])),
    list_to_atom(String).


name_new(P) ->
    make_name("m", P).

temp_new(P) ->
    {var, name_new(P)}.

temp_new([], _) ->
    [];
temp_new([_|L], Pos) ->
    [temp_new(pos:left(Pos)) | temp_new(L, pos:right(Pos))].


