%%% File    : scfa_file.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@it.uu.se>
%%% Purpose : Interface between analyzer and file system
%%% Created :  7 Feb 2000 by Sven-Olof Nystr|m <svenolof@it.uu.se>

-module(scfa_file).
-author('svenolof@it.uu.se').

-export([find_module/3, normalize/1, check_path/1]).

% find_module looks for file with name M and extension Ext. Either
% fails with not_found, or returns {found, FileName}.

%: +type list(X) = erlang:list(X).
%: +type string() = erlang:string().

%: +func find_module :: program:path() * atom() * atom() -> 
%:                     {ok, filename:filename_all()} + not_found.

find_module(Path, Name, Ext) ->
    NameS = atom_to_list(Name),
    ExtS = atom_to_list(Ext),
    find_filename(Path, NameS, "."++ExtS).

find_filename([Dir | Path], Name, Suffix) ->
    Absname = filename:absname([Name,Suffix],Dir),
    case catch file:read_file_info(Absname) of
	{ok,_} -> 
	    {ok, Absname};
	_ ->
	    find_filename(Path, Name, Suffix)
    end;
find_filename([], _, _) ->
    not_found.

%% Normalize a given path.
%% The function check_path catches erroneous paths, thus we assume
%% here that the path is correct.

normalize([P|Path]) when is_atom(P) ->
    normalize([atom_to_list(P)|Path]);
normalize([P|Path]) when is_list(P) ->
    case int_list(P) of
	true -> [filename:join([P])|normalize(Path)];
	_    -> [P|normalize(Path)]
    end;
normalize([P|Path]) ->
    [P|normalize(Path)];
normalize([]) ->
    [];
normalize(Other) ->
    Other.

%: +func check_path :: program:path() -> true + {error, any()}.
check_path([]) -> 
    true;
check_path([Dir |Tail]) ->
    case catch file:read_file_info(Dir) of
	{ok,_} -> 
	    check_path(Tail);
	_ -> 
	    {error, bad_directory}
    end;
check_path(_) ->
    {error, bad_path}.

int_list([H|T]) when is_integer(H) -> int_list(T);
int_list([_|_])                    -> false;
int_list([])                       -> true.


