%%% File    : wl.erl
%%% Author  : Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>
%%% Purpose : 
%%% Created : 16 Feb 1999 by Sven-Olof Nystr|m <svenolof@harpo.csd.uu.se>

-module(worklist).
-author('svenolof@harpo.csd.uu.se').

%: +type worklist(K, W) = {wl, integer(), integer(),
%:                           map(integer(), W), 
%:                           map(K, integer())}.

%: +type list(X) = erlang:list(X).

-export([empty/0, insert/3,  insert_list/3, get_work/1]).

%: +func empty :: () ->  worklist(K, W).

empty() -> 
    {wl, 0, 0, maps:new(), maps:new()}.

% 200126: Erlang maps lack a take-smallest operation, so I need to
% introduce a new representation. {wl, Start, Next, IndexToWork,
% WorkToIndex} uses a set of indices [Start, ..., Next-1]. IndexToWork
% maps indices to work, WorkToIndex maps work to their index.

%: +func insert :: K * W * worklist(K, W) -> worklist(K, W).

insert(Key, Work, {wl, Start, Next, IndexToWork, WorkToIndex}) ->
    case maps:is_key(Key, WorkToIndex) of
	true ->
	    {wl, Start, Next, IndexToWork, WorkToIndex};
	false  ->
	    {wl, Start, Next+1, 
	     maps:put(Next, Work, IndexToWork),
	     maps:put(Key, Next, WorkToIndex)}
    end.

%: +func insert_list :: list(K) * list(W) * worklist(K, W) -> worklist(K, W).

insert_list([Key| KeyList], [Work | List], WL) ->
    insert_list(KeyList, List, insert(Key, Work, WL));
insert_list([], [], WL) ->
    WL.

%: +func get_work :: worklist(K, W) -> none + {value, {W, worklist(K, W)}}.

get_work({wl, Start, Next, IndexToWork, WorkToIndex}) ->
    case Start==Next of
	true ->
	    none;
	false ->
	    Work = maps:get(Start, IndexToWork),
	    {value, {Work, {wl, Start+1, Next, IndexToWork, WorkToIndex}}}
    end.

